/*
 * Prototypes
 */
/*
void setdebug _P((char *, struct socket *));
void setkdebug _P((char *, struct socket *));
void setpassive _P((char *, struct socket *));
void setinitopt _P((char *, struct socket *));
void noopt _P((char *, struct socket *));
void setnovj _P((char *, struct socket *));
void setnovjccomp _P((char *, struct socket *));
void setvjslots _P((char *, struct socket *));
void reqpap _P((char *, struct socket *));
void nopap _P((char *, struct socket *));
void setupapfile _P((char *, struct socket *));
void nochap _P((char *, struct socket *));
void reqchap _P((char *, struct socket *));
void setspeed _P((char *, struct socket *));
void noaccomp _P((char *, struct socket *));
void noasyncmap _P((char *, struct socket *));
void noipaddr _P((char *, struct socket *));
void nomagicnumber _P((char *, struct socket *));
void setasyncmap _P((char *, struct socket *));
void setescape _P((char *, struct socket *));
void setmru _P((char *, struct socket *));
void setmtu _P((char *, struct socket *));
void nomru _P((char *, struct socket *));
void nopcomp _P((char *, struct socket *));
void setconnector _P((char *, struct socket *));
void setdisconnector _P((char *, struct socket *));
void setdomain _P((char *, struct socket *));
void setnetmask _P((char *, struct socket *));
void setcrtscts _P((char *, struct socket *));
void setxonxoff _P((char *, struct socket *));
void setnodetach _P((char *, struct socket *));
void setmodem _P((char *, struct socket *));
void setlocal _P((char *, struct socket *));
void setlock _P((char *, struct socket *));
void setname _P((char *, struct socket *));
void set_user _P((char *, struct socket *));
void setremote _P((char *, struct socket *));
void setauth _P((char *, struct socket *));
void readfile _P((char *, struct socket *));
void setdefaultroute _P((char *, struct socket *));
void setproxyarp _P((char *, struct socket *));
void setpersist _P((char *, struct socket *));
void setdologin _P((char *, struct socket *));
void setusehostname _P((char *, struct socket *));
void setnoipdflt _P((char *, struct socket *));
void setlcptimeout _P((char *, struct socket *));
void setlcpterm _P((char *, struct socket *));
void setlcpconf _P((char *, struct socket *));
void setlcpfails _P((char *, struct socket *));
void setipcptimeout _P((char *, struct socket *));
void setipcpterm _P((char *, struct socket *));
void setipcpconf _P((char *, struct socket *));
void setipcpfails _P((char *, struct socket *));
void setpaptimeout _P((char *, struct socket *));
void setpapreqs _P((char *, struct socket *));
void setchaptimeout _P((char *, struct socket *));
void setchapchal _P((char *, struct socket *));
void setchapintv _P((char *, struct socket *));
void setipcpaccl _P((char *, struct socket *));
void setipcpaccr _P((char *, struct socket *));
void setlcpechointv _P((char *, struct socket *));
void setlcpechofails _P((char *, struct socket *));

*/
